/*
 *	This is a part of adsbox - ADS-B decoder software
 * 	(c) atty 2011-13 romulmsg@mail.ru
 *	Alexander Heidrich - DF4, Flight Status decode
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "misc.h"
#include "../sqlite3/sqlite3.h"
#include "db.h"

#ifndef round
double round(double value) {
	if (value < 0)
		return -(floor(-value + 0.5));
	else
		return floor(value + 0.5);
}
#endif

#define MAXDIST 440.0 // reject too far position
#define MAXSTEP 20.0 // reject strange position
char cs_tbl[] = { '@', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K',
		'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y',
		'Z', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ',
		' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', '0', '1', '2', '3', '4', '5',
		'6', '7', '8', '9', ' ', ' ', ' ', ' ', ' ', ' ' };

extern double homelat;
extern double homelon;

double dist(double lat, double lon, double lat1, double lon1) {
	return (6371 * acos(sin(lat * M_PI / 180.) * sin(lat1 * M_PI / 180.) + cos(
			lat * M_PI / 180.) * cos(lat1 * M_PI / 180.) * cos((lon - lon1)
			* M_PI / 180.)));
}

void azimuth(double *rlat1, double *rlon1, double *rlat2, double *rlon2, double *a) {
	double lat1, lon1, lat2, lon2, b;
	double result = 0;

	lat1 = *rlat1 * M_PI / 180;
	lon1 = *rlon1 * M_PI / 180;
	lat2 = *rlat2 * M_PI / 180;
	lon2 = *rlon2 * M_PI / 180;

	b = acos(sinf(lat2) * sin(lat1) + cos(lat2) * cos(lat1) * cos(lon2 - lon1));
	result = asinf(cos(lat2) * sin(lon2 - lon1) / sin(b));

	if ((lat2 > lat1) && (lon2 > lon1)) {
	} else if ((lat2 < lat1) && (lon2 < lon1)) {
		result = M_PI - result;
	} else if ((lat2 < lat1) && (lon2 > lon1)) {
		result = M_PI - result;
	} else if ((lat2 > lat1) && (lon2 < lon1)) {
		result = result + 2 * M_PI;
	}

	result = result * 180 / M_PI;
	*a = result;
}

int NL(double lat) {
	if (fabs(lat) < 10.47047130)
		return (59);
	else if (fabs(lat) < 14.82817437)
		return (58);
	else if (fabs(lat) < 18.18626357)
		return (57);
	else if (fabs(lat) < 21.02939493)
		return (56);
	else if (fabs(lat) < 23.54504487)
		return (55);
	else if (fabs(lat) < 25.82924707)
		return (54);
	else if (fabs(lat) < 27.93898710)
		return (53);
	else if (fabs(lat) < 29.91135686)
		return (52);
	else if (fabs(lat) < 31.77209708)
		return (51);
	else if (fabs(lat) < 33.53993436)
		return (50);
	else if (fabs(lat) < 35.22899598)
		return (49);
	else if (fabs(lat) < 36.85025108)
		return (48);
	else if (fabs(lat) < 38.41241892)
		return (47);
	else if (fabs(lat) < 39.92256684)
		return (46);
	else if (fabs(lat) < 41.38651832)
		return (45);
	else if (fabs(lat) < 42.80914012)
		return (44);
	else if (fabs(lat) < 44.19454951)
		return (43);
	else if (fabs(lat) < 45.54626723)
		return (42);
	else if (fabs(lat) < 46.86733252)
		return (41);
	else if (fabs(lat) < 48.16039128)
		return (40);
	else if (fabs(lat) < 49.42776439)
		return (39);
	else if (fabs(lat) < 50.67150166)
		return (38);
	else if (fabs(lat) < 51.89342469)
		return (37);
	else if (fabs(lat) < 53.09516153)
		return (36);
	else if (fabs(lat) < 54.27817472)
		return (35);
	else if (fabs(lat) < 55.44378444)
		return (34);
	else if (fabs(lat) < 56.59318756)
		return (33);
	else if (fabs(lat) < 57.72747354)
		return (32);
	else if (fabs(lat) < 58.84763776)
		return (31);
	else if (fabs(lat) < 59.95459277)
		return (30);
	else if (fabs(lat) < 61.04917774)
		return (29);
	else if (fabs(lat) < 62.13216659)
		return (28);
	else if (fabs(lat) < 63.20427479)
		return (27);
	else if (fabs(lat) < 64.26616523)
		return (26);
	else if (fabs(lat) < 65.31845310)
		return (25);
	else if (fabs(lat) < 66.36171008)
		return (24);
	else if (fabs(lat) < 67.39646774)
		return (23);
	else if (fabs(lat) < 68.42322022)
		return (22);
	else if (fabs(lat) < 69.44242631)
		return (21);
	else if (fabs(lat) < 70.45451075)
		return (20);
	else if (fabs(lat) < 71.45986473)
		return (19);
	else if (fabs(lat) < 72.45884545)
		return (18);
	else if (fabs(lat) < 73.45177442)
		return (17);
	else if (fabs(lat) < 74.43893416)
		return (16);
	else if (fabs(lat) < 75.42056257)
		return (15);
	else if (fabs(lat) < 76.39684391)
		return (14);
	else if (fabs(lat) < 77.36789461)
		return (13);
	else if (fabs(lat) < 78.33374083)
		return (12);
	else if (fabs(lat) < 79.29428225)
		return (11);
	else if (fabs(lat) < 80.24923213)
		return (10);
	else if (fabs(lat) < 81.19801349)
		return (9);
	else if (fabs(lat) < 82.13956981)
		return (8);
	else if (fabs(lat) < 83.07199445)
		return (7);
	else if (fabs(lat) < 83.99173563)
		return (6);
	else if (fabs(lat) < 84.89166191)
		return (5);
	else if (fabs(lat) < 85.75541621)
		return (4);
	else if (fabs(lat) < 86.53536998)
		return (3);
	else if (fabs(lat) < 87.00000000)
		return (2);
	else
		return (1);
}

double modulo(double x, double y) {
	double r;
	r = x - y * floor(x / y);
	return (r);
}

int crc(char * buf) {
	int poly;
	int data, data1, data2;
	int i;

	poly = 0xfffa0480;
	sscanf(buf, "*%08x%08x%08x", &data, &data1, &data2);
	data2 = data2 & 0xffffff00;
	for (i = 0; i < 88; i++) {
		if ((data & 0x80000000) != 0)
			data = data ^ poly;
		data = data << 1;
		if ((data1 & 0x80000000) != 0)
			data = data | 1;
		data1 = data1 << 1;
		if ((data2 & 0x80000000) != 0)
			data1 = data1 | 1;
		data2 = data2 << 1;
	}
	return (data);
}

int crc56(char * buf) {
	int poly;
	int data;
	int i;

	poly = 0xfffa0480;
	sscanf(buf, "*%08x", &data);
	for (i = 0; i < 32; i++) {
		if ((data & 0x80000000) != 0)
			data = data ^ poly;
		data = data << 1;
	}
	return (data);
}

void flight_status(unsigned char *data, int *ground, int *alert, int *spi) {
	char fs = *data & 0x07;
	//DEBUG("FS:%d\n",fs);
	if (fs == 0) {
		*ground = 0;
		*alert = 0;
		*spi = 0;
	} // airborne
	else if (fs == 1) {
		*ground = -1;
		*alert = 0;
		*spi = 0;
	} // ground
	else if (fs == 2) {
		*ground = 0;
		*alert = -1;
		*spi = 0;
	} // airborne + squawk change
	else if (fs == 3) {
		*ground = -1;
		*alert = -1;
		*spi = 0;
	} // ground + squawk change
	else if (fs == 4) {
		*ground = 0;
		*alert = -1;
		*spi = -1;
	} else if (fs == 5) {
		*ground = 0;
		*alert = 0, *spi = -1;
	} // airborne or ground, spi (see fs == 4)
}

unsigned int find_squawk(unsigned char *data2, unsigned char *data3,
		int *emergency) {
	int squawk = (((*data3 >> 7) & 0x01) << 11)
			| (((*data2 >> 1) & 0x01) << 10) | (((*data2 >> 3) & 0x01) << 9) |

	(((*data3 >> 1) & 0x01) << 8) | (((*data3 >> 3) & 0x01) << 7) | (((*data3
			>> 5) & 0x01) << 6) |

	((*data2 & 0x01) << 5) | (((*data2 >> 2) & 0x01) << 4) | (((*data2 >> 4)
			& 0x01) << 3) |

	((*data3 & 0x01) << 2) | (((*data3 >> 2) & 0x01) << 1) | (((*data3 >> 4)
			& 0x01));

	if (squawk == 07500 || squawk == 07600 || squawk == 07700) {
		*emergency = -1;
	} else {
		*emergency = 0;
	}
	return squawk;
}

int ac_alt_decode(unsigned char *data2, unsigned char *data3) {
	// port of Spruts code from Delphi to C
	unsigned char a1, a2, a4, b1, b2, b4, c1, c2, c4, d2, d4, m1, q1;
	int alt;
	int alt_code;
	char alt_grey;
	char subcode;
	char altitude_grey_sub;
	int copybit;
	int k;
	int result_grey_decode;
	char maske;

	a4 = ((*data3 >> 7) & 0x01);
	a2 = ((*data2 >> 1) & 0x01);
	a1 = ((*data2 >> 3) & 0x01);

	b4 = ((*data3 >> 1) & 0x01);
	b2 = ((*data3 >> 3) & 0x01);
	b1 = ((*data3 >> 5) & 0x01);

	c4 = (*data2 & 0x01);
	c2 = ((*data2 >> 2) & 0x01);
	c1 = ((*data2 >> 4) & 0x01);

	d4 = (*data3 & 0x01);
	d2 = ((*data3 >> 2) & 0x01);

	//DEBUG("A4:%d, A2:%d, A1:%d, B4:%d, B2:%d, B1:%d, C4:%d, C2:%d, C1:%d, D4:%d, D2:%d\n", a4, a2, a1, b4, b2, b1, c4,c2,c1, d4,d2);         

	m1 = ((*data3 >> 6) & 0x01);
	q1 = ((*data3 >> 4) & 0x01);

	if (q1) {
		alt = (c1 * 1024 + a1 * 512 + c2 * 256 + a2 * 128 + c4 * 64 + a4 * 32
				+ b1 * 16 + b2 * 8 + d2 * 4 + b4 * 2 + d4) * 25 - 1000;
		if (m1) {
			alt = round((alt + 1000.0) / 25.0);
			alt = alt * 10 - 300;
			alt = round(alt * 3.281);
		}
	} else {
		alt_code = (d2 * 1024 + d4 * 512 + a1 * 256 + a2 * 128 + a4 * 64 + b1
				* 32 + b2 * 16 + b4 * 8 + c1 * 4 + c2 * 2 + c4);
		alt_grey = alt_code >> 3;
		subcode = alt_code & 0x07;
		altitude_grey_sub = 0;
		copybit = 0;
		k = 0;
		result_grey_decode = 0;
		maske = 0x80;

		for (k = 0; k < 16; k++) {
			if (alt_grey & maske != 0) {
				copybit = !copybit;
			}
			if (copybit) {
				result_grey_decode = result_grey_decode | maske;
			}
			maske >> 1;
		}

		if (result_grey_decode % 2 == 0) {
			switch (subcode) {
			case 1:
				altitude_grey_sub = 0;
			case 3:
				altitude_grey_sub = 1;
			case 2:
				altitude_grey_sub = 2;
			case 6:
				altitude_grey_sub = 3;
			case 4:
				altitude_grey_sub = 4;
			}
		} else {
			switch (subcode) {
			case 4:
				altitude_grey_sub = 0;
			case 6:
				altitude_grey_sub = 1;
			case 2:
				altitude_grey_sub = 2;
			case 3:
				altitude_grey_sub = 3;
			case 1:
				altitude_grey_sub = 4;
			}
		}
		alt = result_grey_decode * 500 + altitude_grey_sub * 100 - 1200;
	}
	return alt;
}

int decode_message56(char * buf, char * msg) {
	unsigned char rawdata[7];
	unsigned char df;
	int c;
	int alt;
	int emergency = 0;
	int spi = 0;
	int alert = 0;
	int ground = 0;
	unsigned int icao;
	unsigned int squawk;

	sscanf(buf, "*%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx", &rawdata[0],
			&rawdata[1], &rawdata[2], &rawdata[3], &rawdata[4], &rawdata[5],
			&rawdata[6]);

	df = (rawdata[0] >> 3) & 0x1F;
	DEBUG("\nDF:%u\t", df);

	if (df == 0) {
		icao = ((crc56(buf) >> 8) & 0xffffff) ^ ((rawdata[4] << 16)
				| (rawdata[5] << 8) | rawdata[6]);
		alt = ac_alt_decode(&rawdata[2], &rawdata[3]);
		if (db_update_alt(&icao, &alt) == 1) {
			sprintf(msg, "MSG,5,,,%06X,,,,,,,%d,,,,,,,,,,\n\0", icao, alt);
			DEBUG ("ICAO:%06x Alt:%d", icao, alt);
		} else
			return 0;
	}

	if (df == 4) {
		icao = ((crc56(buf) >> 8) & 0xffffff) ^ ((rawdata[4] << 16)
				| (rawdata[5] << 8) | rawdata[6]);
		alt = ac_alt_decode(&rawdata[2], &rawdata[3]);
		if (db_update_alt(&icao, &alt) == 1) {
			flight_status(&rawdata[0], &ground, &alert, &spi);
			sprintf(msg, "MSG,5,,,%06X,,,,,,,%d,,,,,,,%d,%d,%d,%d\n\0", icao,
					alt, alert, emergency, spi, ground);
			DEBUG ("ICAO:%06x Alt:%d Alert:%d Emergency:%d SPI:%d Ground:%d", icao, alt, alert, emergency, spi, ground);
		} else
			return 0;
	}
	if (df == 5) { // surveillance identify reply
		icao = ((crc56(buf) >> 8) & 0xffffff) ^ ((rawdata[4] << 16)
				| (rawdata[5] << 8) | rawdata[6]);
		squawk = find_squawk(&rawdata[2], &rawdata[3], &emergency);
		if (db_update_squawk(&icao, &squawk) == 1) {
			flight_status(&rawdata[0], &ground, &alert, &spi);
			sprintf(msg, "MSG,6,,,%06X,,,,,,,,,,,,,%04o,%d,%d,%d,%d\n\0", icao,
					squawk, alert, emergency, spi, ground);
			DEBUG ("ICAO:%06x Squawk:%04o Alert:%d Emergency:%d SPI:%d Ground:%d", icao, squawk, alert, emergency, spi, ground);
		} else
			return 0;
	}
	if (df == 11) { // all call reply address announce
			c = crc56(buf);
			if (c != ((rawdata[4] << 24) | (rawdata[5] << 16) | rawdata[6] << 8)) {
				DEBUG("Bad CRC !");
				return (0);
			}
			icao = (rawdata[1] << 16) | (rawdata[2] << 8) | rawdata[3];
			db_update_icao(&icao);
			DEBUG("ICAO:%06x", icao);
			return (0);
	}
	return strlen(msg);
}

int decode_message(char * buf, char * msg) {
	unsigned char rawdata[14];
	unsigned char df;
	unsigned char tc;
	unsigned int af; // Application field
	char callsign[9];
	unsigned int icao;
	unsigned int cs;
	unsigned char f;
	int lat, lon;
	int alt;
	unsigned char i;
	int j, ni, m;
	double rlat0, rlat1;
	int NL0, NL1;
	double dlon;
	double rlat, rlon;
	double d;
	int c;
	unsigned char subtype;
	double ew_spd, ns_spd;
	float gnd_spd, trk;
	int vr;
	unsigned char ew_dir, ns_dir;
	unsigned int squawk;
	int emergency = 0;
	int spi = 0;
	int alert = 0;
	int ground = 0;
	int alt_code;
	cpr_info ci;

	sscanf(
			buf,
			"*%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx",
			&rawdata[0], &rawdata[1], &rawdata[2], &rawdata[3], &rawdata[4],
			&rawdata[5], &rawdata[6], &rawdata[7], &rawdata[8], &rawdata[9],
			&rawdata[10], &rawdata[11], &rawdata[12], &rawdata[13]);

	df = (rawdata[0] >> 3) & 0x1F;
	DEBUG("\nDF:%u\t", df);

	if (df == 17) {
		c = crc(buf);
		if (c != ((rawdata[11] << 24) | (rawdata[12] << 16) | rawdata[13] << 8)) {
			DEBUG("Bad CRC !");
			return (0);
		}
		icao = (rawdata[1] << 16) | (rawdata[2] << 8) | rawdata[3];
		tc = (rawdata[4] >> 3) & 0x1f;
		//DEBUG("ICAO:%06x TC:%u\n", icao, tc);

		if (tc == 4) { // aircraft ID
			bzero(&callsign, 9);
			cs = (rawdata[5] << 16) | (rawdata[6] << 8) | (rawdata[7]);
			callsign[0] = cs_tbl[(cs >> 18) & 0x3f];
			callsign[1] = cs_tbl[(cs >> 12) & 0x3f];
			callsign[2] = cs_tbl[(cs >> 6) & 0x3f];
			callsign[3] = cs_tbl[cs & 0x3f];
			cs = (rawdata[8] << 16) | (rawdata[9] << 8) | rawdata[10];
			callsign[4] = cs_tbl[(cs >> 18) & 0x3f];
			callsign[5] = cs_tbl[(cs >> 12) & 0x3f];
			callsign[6] = cs_tbl[(cs >> 6) & 0x3f];
			callsign[7] = cs_tbl[cs & 0x3f];

			db_update_callsign(&icao, callsign);
			sprintf(msg, "MSG,1,,,%06X,,,,,,%s,,,,,,,,0,0,0,0\n\0", icao,
					callsign);
			DEBUG ("ICAO:%06x Callsign:%s", icao, callsign);
		}

		else if ((tc > 8) && (tc < 19)) { // aircraft air position
			f = (rawdata[6] >> 2) & 0x01;
			lat = ((rawdata[8] >> 1) & 0x7f) | (rawdata[7] << 7) | ((rawdata[6]
					& 0x03) << 15);
			lon = rawdata[10] | (rawdata[9] << 8) | ((rawdata[8] & 0x01) << 16);
			alt = 25 * (((rawdata[6] >> 4) & 0x0f)
					| (((rawdata[5] >> 1) & 0x7f) << 4)) - 1000;

			ci = db_update_cpr(&icao, &f, &lat, &lon);

			if ((ci.lat0 == 0) || (ci.lat1 == 0))
				return 0; //no odd and even both

			j = floor((59. * ci.lat0 - 60. * ci.lat1) / 131072. + 0.5);

			rlat0 = 6. * (modulo(j, 60.) + ci.lat0 / 131072.);
			rlat1 = 6.101694915254237288 * (modulo(j, 59.) + ci.lat1 / 131072.);

			rlat0 = (rlat0 < 270 ? rlat0 : rlat0 - 360.);
			rlat1 = (rlat1 < 270 ? rlat1 : rlat1 - 360.);

			NL0 = NL(rlat0);
			NL1 = NL(rlat1);

			if (NL0 != NL1) 
				return 0;

			m = floor((ci.lon0 * (NL0 - 1) - ci.lon1 * NL1) / 131072. + 0.5);

			if (f == 0) {
				ni = (NL0 > 1 ? NL0 : 1);
				dlon = 360. / (double) ni;
				rlat = rlat0;
				rlon = dlon * (modulo(m, ni) + (double) ci.lon0 / 131072.);
			} else {
				ni = ((NL1 - 1) > 1 ? (NL1 - 1) : 1);
				dlon = 360. / (double) ni;
				rlat = rlat1;
				rlon = dlon * (modulo(m, ni) + (double) ci.lon1 / 131072.);
			}

			rlon = (rlon > 180 ? rlon - 360. : rlon); // western long.
			
			if ((homelat != 0) && (homelon != 0)) {
				d = dist(rlat, rlon, homelat, homelon);
				DEBUG("Distance %f ", d);
				if (d > MAXDIST)
					return (0); //reject too far position
			}
			if ((ci.lat != 0) && (ci.lon != 0) && (dist(rlat, rlon,
			    ci.lat, ci.lon) > MAXSTEP)) 
			    return (0); // reject too large track step

			db_update_position(&icao, &rlat, &rlon);
			db_update_alt(&icao, &alt);
			//note: MSG3 would also require the FS_status fields
			//to be populated with a value. In DF17 the corresponding
			//flags can not be "calculated" because they are not transmitted.
			sprintf(msg, "MSG,3,,,%06X,,,,,,,%d,,,%1.5f,%1.5f,,,0,0,0,0\n\0",
					icao, alt, rlat, rlon);
			DEBUG ("ICAO:%06x Alt:%d Lat:%1.5f Lon:%1.5f", icao, alt, rlat, rlon);

		} else if (tc == 19) { // air speed
			subtype = rawdata[4] & 0x07;
			if (subtype == 1) { // ground non supersonic speed
				ew_spd = rawdata[6] | ((rawdata[5] & 0x03) << 8); // -1 TODO
				ns_spd = ((rawdata[8] >> 5) & 0x07)
						| ((rawdata[7] & 0x7f) << 3);
				if ((ew_spd == 0) && (ns_spd == 0))
					return (0);

				ew_dir = (rawdata[5] >> 2) & 0x01;
				ns_dir = (rawdata[7] >> 7) & 0x01;
				gnd_spd = floor(sqrt(ew_spd * ew_spd + ns_spd * ns_spd));
				if ((!ew_dir) && (!ns_dir))
					trk = (ew_spd == 0 ? 0 : 90 - 180. / M_PI * atan(ns_spd
							/ ew_spd));
				if ((!ew_dir) && (ns_dir))
					trk = (ew_spd == 0 ? 180 : 90 + 180. / M_PI * atan(ns_spd
							/ ew_spd));
				if ((ew_dir) && (ns_dir))
					trk = (ew_spd == 0 ? 180 : 270 - 180. / M_PI * atan(ns_spd
							/ ew_spd));
				if ((ew_dir) && (!ns_dir))
					trk = (ew_spd == 0 ? 0 : 270 + 180. / M_PI * atan(ns_spd
							/ ew_spd));
				vr = ((rawdata[8] & 0x08) == 0 ? 1 : -1) * ((((rawdata[9] >> 2)
						& 0x3f) | (rawdata[8] & 0x07) << 6) - 1) * 64;
				db_update_speed_heading(&icao, &gnd_spd, &trk, &vr);
				sprintf(msg,
						"MSG,4,,,%06X,,,,,,,,%1.1f,%1.1f,,,%i,,0,0,0,0\n\0",
						icao, gnd_spd, trk, vr);
				DEBUG ("ICAO:%06x Speed:%1.1f Track:%1.1f Vertical:%i", icao, gnd_spd, trk, vr);
			} else if (subtype == 3) { // air non supersonic speed
			    trk = (rawdata[6] | ((rawdata[5] & 0x03) << 8)) * 360.0 / 1024.0;
			    gnd_spd = ((rawdata[8] >> 5) & 0x07) | ((rawdata[7] & 0x7f) << 3);
			    if (gnd_spd == 0)
				return (0);
			    vr = ((rawdata[8] & 0x08) == 0 ? 1 : -1) * ((((rawdata[9] >> 2)
				& 0x3f) | (rawdata[8] & 0x07) << 6) - 1) * 64;
			    db_update_speed_heading(&icao, &gnd_spd, &trk, &vr);
			    sprintf(msg,
				"MSG,4,,,%06X,,,,,,,,%1.1f,%1.1f,,,%i,,0,0,0,0\n\0",
				icao, gnd_spd, trk, vr);
			    DEBUG ("ICAO:%06x Speed:%1.1f Track:%1.1f Vertical:%i", icao, gnd_spd, trk, vr);
			}
		} else
		DEBUG("ICAO:%06x TC:%u not implemented yet.", icao, tc);
	}
	/*if (df == 19) {
		af = rawdata[0] & 0x07;
		icao = (rawdata[1] << 16) | (rawdata[2] << 8) | rawdata[3];
		tc = (rawdata[4] >> 3) & 0x1f;
		if (af == 0)
			DEBUG ("ICAO:%06x AF:%u TC:%u", icao, af, tc);
	}*/

	if (df == 20) { // COMM-B altitude reply
		icao = ((crc(buf) >> 8) & 0xffffff) ^ ((rawdata[11] << 16) | (rawdata[12]
						<< 8) | rawdata[13]);
		alt = ac_alt_decode(&rawdata[2], &rawdata[3]);
		if (db_update_alt(&icao, &alt) == 1) {
			sprintf(msg, "MSG,5,,,%06X,,,,,,,%d,,,,,,,,,,\n\0", icao, alt);
			DEBUG ("ICAO:%06x Alt:%d", icao, alt);
		} else
			return 0;
	}

	if (df == 21) { //COMM-B identity reply
		icao = ((crc(buf) >> 8) & 0xffffff) ^ ((rawdata[11] << 16) | (rawdata[12]
						<< 8) | rawdata[13]);
		squawk = find_squawk(&rawdata[2], &rawdata[3], &alert);
		if (db_update_squawk(&icao, &squawk) == 1) {
			flight_status(&rawdata[0], &ground, &alert, &spi);
			sprintf(msg, "MSG,6,,,%06X,,,,,,,,,,,,,%04o,%d,%d,%d,%d\n\0", icao,
					squawk, alert, emergency, spi, ground);
			DEBUG ("ICAO:%06x Squawk:%04o Alert:%d, Emergency:%d, SPI:%d, Ground:%d", icao, squawk, alert, emergency, spi, ground);
		}
	}
	return strlen(msg);
}
