/*
 *	This is a part of adsbox - ADS-B decoder software
 * 	(c) atty 2011-12 romulmsg@mail.ru
 */

#include <stdio.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <pthread.h>
#include <string.h>
#include <strings.h>
#include "misc.h"
#include "../sqlite3/sqlite3.h"

#define REQBUFLEN 1024

extern sqlite3 *db;
//extern sqlite3_mutex *db_mux;
extern double homelat;
extern double homelon;

void http_header(int *s, int status, char *title, char *mime, int length) {
	char buf[128];

	sprintf(buf, "HTTP/1.0 %i %s\r\n\0", status, title);
	send(*s, buf, strlen(buf), 0);
	sprintf(buf, "Server: ADSBox ver.%s\r\n\0", VERSION);
	send(*s, buf, strlen(buf), 0);
	sprintf(buf, "Content-Type: %s\r\n\0", mime);
	send(*s, buf, strlen(buf), 0);
	if (length > -1) {
		sprintf(buf, "Content-Length: %i\r\n\0", length);
		send(*s, buf, strlen(buf), 0);
	}
	sprintf(buf, "Connection: close\r\n\r\n\0");
	send(*s, buf, strlen(buf), 0);
}

void http_error(int *s, int status, char *title, char *info) {
	char buf[128];

	http_header((int *) s, status, title, "text/html", -1);

	sprintf(buf, "<html>\n<head><title>%i %s</title></head>\n<body>\n\0",
			status, title);
	send(*s, buf, strlen(buf), 0);
	sprintf(buf,
			"<h3>%i %s</h3><br>%s<hr><i>ADSBox ver.%s http server.</i>\n\0",
			status, title, info, VERSION);
	send(*s, buf, strlen(buf), 0);
	sprintf(buf, "</body>\n</html>\n\0");
	send(*s, buf, strlen(buf), 0);
}

int list_flights(int *s) {
	char buf[REQBUFLEN];
	char ** results;
	int rc, rows, cols, i;
	char *zErrMsg = 0;
	char *sql = malloc(256);
	sqlite3_stmt *stmt;

	http_header(s, 200, "OK", "text/html", -1);

	sprintf(
			buf,
			"<html>\n<body>\n\
			<div align=\"center\"><h4>Flight board.</h4></div>\
			<div align=\"center\">\n<table border=\"0\" cellpadding=\"4\" cellspacing=\"1\">\
			<tr bgcolor=\"#AAAAFF\">\n\0");
	send(*s, buf, strlen(buf), 0);

	sprintf(
			buf,
			"<td>ICAO</td><td>LAT</td><td>LON</td><td>ALT</td>\
			<td>FLIGHT</td><td>&nbsp;</td><td>SPEED</td><td>VSPEED</td>\
			<td>HEADING</td><td>SQUAWK</td><td>SEEN_TIME</td></tr>\n\0");
	send(*s, buf, strlen(buf), 0);

	//sqlite3_mutex_enter(db_mux);
	sprintf(
			sql,
			"SELECT ICAO, LAT, LON, ALT, CALLSIGN, SPEED, VSPEED, HEADING, SQUAWK, \
			DATETIME(SEEN_TIME, 'UNIXEPOCH', 'LOCALTIME') \
			FROM FLIGHTS\0");
	rc = sqlite3_prepare(db, sql, -1, &stmt, 0);
	if (rc != SQLITE_OK) {
		DEBUG("SQLite error: could not prepare.");
		free(sql);
		//sqlite3_mutex_leave(db_mux);
		return 1;
	}

	while (sqlite3_step(stmt) == SQLITE_ROW) {
		sprintf(
				buf,
				"<tr><td>%06X</td><td>%1.3f</td><td>%1.3f</td><td>%s</td><td>%s</td>\
				<td>%s</td><td>%s</td><td>%i</td><td>%1.0f&deg;</td><td>%04o</td><td>%s</td></tr>\n\0",
				sqlite3_column_int(stmt, 0),
				sqlite3_column_double(stmt, 1),
				sqlite3_column_double(stmt, 2),
				sqlite3_column_text(stmt, 3),
				sqlite3_column_text(stmt, 4),
				(sqlite3_column_int(stmt, 6) > 128 ? "<img src=\"depart.png\">"
						: (sqlite3_column_int(stmt, 6) < -128 ? "<img src=\"arrive.png\">"
								: "<img src=\"flight.png\">")),
				sqlite3_column_text(stmt, 5), sqlite3_column_int(stmt, 6),
				sqlite3_column_double(stmt, 7), sqlite3_column_int(stmt, 8),
				sqlite3_column_text(stmt, 9));
		send(*s, buf, strlen(buf), 0);
	}
	sqlite3_finalize(stmt);

	sprintf(buf, "</table>\n</div>\n<br><pre>\n\0");
	send(*s, buf, strlen(buf), 0);

	/*sprintf(
			sql,
			"SELECT AZIMUTH, LAT, LON, DIST FROM RADAR_RANGE\0");
	rc = sqlite3_get_table(db, sql, &results, &rows, &cols, &zErrMsg);
	if (rc != SQLITE_OK) {
		DEBUG("SQLite error: %s\n", zErrMsg);
		sqlite3_free(zErrMsg);
		sqlite3_free_table(results);
		return 1;
	}

	for (i = 0; i <= rows; i++) {
		sprintf(buf, "%s\t%s\t%s\t%s\n\0", results[i * cols], results[i
				* cols + 1], results[i * cols + 2], results[i * cols + 3],
				results[i * cols + 4]);
		send(*s, buf, strlen(buf), 0);
	}*/

	sprintf(buf, "</pre></body>\n</html>\n\0");
	send(*s, buf, strlen(buf), 0);

	//sqlite3_free_table(results);
	free(sql);
	//sqlite3_mutex_leave(db_mux);
	return 0;
}

int make_kml(int *s) {
	char * buf = malloc(1024);
	char *sql_f = malloc(256); // flights
	sqlite3_stmt *stmt_f;

	char *sql = malloc(256); // tracks
	sqlite3_stmt *stmt;

	http_header(s, 200, "OK", "text/application/vnd.google-earth.kml+xml", -1);

	strcpy(buf, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n\0");
	send(*s, buf, strlen(buf), 0);
	strcpy(buf,
			"<kml xmlns=\"http://www.opengis.net/kml/2.2\">\n<Document>\n\0");
	send(*s, buf, strlen(buf), 0);
	sprintf(
			buf,
			"<Style id=\"steady\"><IconStyle><Icon><href>steady.png</href></Icon></IconStyle></Style>\n\0");
	send(*s, buf, strlen(buf), 0);
	sprintf(
			buf,
			"<Style id=\"climb\"><IconStyle><Icon><href>climb.png</href></Icon></IconStyle></Style>\n\0");
	send(*s, buf, strlen(buf), 0);
	sprintf(
			buf,
			"<Style id=\"desc\"><IconStyle><Icon><href>desc.png</href></Icon></IconStyle></Style>\n\0");
	send(*s, buf, strlen(buf), 0);
	sprintf(
			buf,
			"<Style id=\"track\"><LineStyle><color>9f7fffff</color><width>1</width></LineStyle>");
	send(*s, buf, strlen(buf), 0);
	sprintf(buf, "<PolyStyle><color>#afD20F0F</color></PolyStyle></Style>");
	send(*s, buf, strlen(buf), 0);

	sprintf(
			buf,
			"<Style id=\"radar\"><LineStyle><color>9f7fffff</color><width>1</width></LineStyle>");
	send(*s, buf, strlen(buf), 0);
	sprintf(buf, "<PolyStyle><color>#5f0fff0F</color></PolyStyle></Style>");
	send(*s, buf, strlen(buf), 0);


	if ((homelat != 0) && (homelon != 0)) {
		sprintf(buf, "<Placemark id=\"0001\">\n<name>Home</name>\n\0");
		send(*s, buf, strlen(buf), 0);
		sprintf(buf,
				"<description>This is your home location.</description>\n\0");
		send(*s, buf, strlen(buf), 0);
		sprintf(buf, "<Point><coordinates>%f,%f,0</coordinates></Point>\n\0",
				homelon, homelat);
		send(*s, buf, strlen(buf), 0);
		sprintf(buf, "</Placemark>\n\0");
		send(*s, buf, strlen(buf), 0);
	}
	//sqlite3_mutex_enter(db_mux);

	sprintf(
			sql_f,
			"SELECT ICAO, LAT, LON, ALT, CALLSIGN, SPEED, HEADING, SQUAWK, VSPEED FROM FLIGHTS \
			WHERE ALT<>0 AND LAT<>0 AND LON<>0\0");
	sqlite3_prepare(db, sql_f, -1, &stmt_f, 0);
	while (sqlite3_step(stmt_f) == SQLITE_ROW) {

		sprintf(buf, "<Placemark id=\"%06X\">\n\0", sqlite3_column_int(stmt_f,
				0));
		send(*s, buf, strlen(buf), 0);
		if (!strlen(sqlite3_column_text(stmt_f, 4)))
			sprintf(buf, "<name>%06X</name>\n\0", sqlite3_column_int(stmt_f, 0));
		else
			sprintf(buf, "<name>%s</name>\n\0", sqlite3_column_text(stmt_f, 4));
		send(*s, buf, strlen(buf), 0);
		sprintf(
				buf,
				"<description><![CDATA[\
							Flight: %s<br>\n\
							ICAO hex: %06X<br>\n\
							Altitude: %s m<br>\n\
							Heading: %1.0f&deg;<br>\n\
							Speed: %1s<br>\n\
							Squawk: %6o<br>\n\
							Lat: %1.2f Lon: %1.2f<br>\n\
							]]></description>\n\0",
				sqlite3_column_text(stmt_f, 4), sqlite3_column_int(stmt_f, 0),
				sqlite3_column_text(stmt_f, 3),
				sqlite3_column_double(stmt_f, 6),
				sqlite3_column_text(stmt_f, 5), sqlite3_column_int(stmt_f, 7),
				sqlite3_column_double(stmt_f, 1), sqlite3_column_double(stmt_f,
						2));
		send(*s, buf, strlen(buf), 0);

		// colour icons
		if (sqlite3_column_int(stmt_f, 8) > 128)
			sprintf(buf, "<styleUrl>#climb</styleUrl>\n\0");
		else if (sqlite3_column_int(stmt_f, 8) < -128)
			sprintf(buf, "<styleUrl>#desc</styleUrl>\n\0");
		else
			sprintf(buf, "<styleUrl>#steady</styleUrl>\n\0");
		send(*s, buf, strlen(buf), 0);

		strcpy(buf, "<Point>\n\0");
		send(*s, buf, strlen(buf), 0);
		strcpy(buf, "<extrude>1</extrude>\n\0");
		send(*s, buf, strlen(buf), 0);
		strcpy(buf, "<altitudeMode>relativeToGround</altitudeMode>\n\0");
		send(*s, buf, strlen(buf), 0);
		sprintf(buf, "<coordinates>%s,%s,%s</coordinates>\n\0",
				sqlite3_column_text(stmt_f, 2), sqlite3_column_text(stmt_f, 1),
				sqlite3_column_text(stmt_f, 3));
		send(*s, buf, strlen(buf), 0);
		strcpy(buf, "</Point>\n</Placemark>\n\0");
		send(*s, buf, strlen(buf), 0);

		strcpy(
				buf,
				"<Placemark>\n<styleUrl>#track</styleUrl>\n<LineString>\n<extrude>1</extrude>\n<tessellate>1</tessellate>\n\0");
		send(*s, buf, strlen(buf), 0);
		strcpy(buf, "<altitudeMode>absolute</altitudeMode>\n<coordinates>\n\0");
		send(*s, buf, strlen(buf), 0);

		sprintf(
				sql,
				"SELECT LON, LAT, ALT FROM TRACKS WHERE ICAO=%s ORDER BY SEEN_TIME DESC\0",
				sqlite3_column_text(stmt_f, 0));
		sqlite3_prepare(db, sql, -1, &stmt, 0);
		while (sqlite3_step(stmt) == SQLITE_ROW) {
			sprintf(buf, "%s,%s,%s\n\0", sqlite3_column_text(stmt, 0),
					sqlite3_column_text(stmt, 1), sqlite3_column_text(stmt, 2));
			send(*s, buf, strlen(buf), 0);
		}
		sqlite3_finalize(stmt);

		strcpy(buf, "</coordinates>\n</LineString>\n\0");
		send(*s, buf, strlen(buf), 0);
		strcpy(buf, "</Placemark>\n\0");
		send(*s, buf, strlen(buf), 0);
	}

	/*
	 *  Radar area
	 */
	if (homelat != 0 && homelon != 0) {
		sprintf(buf,
				"<Placemark>\n<name>Your radar coverage area.</name>\n\0");
		send(*s, buf, strlen(buf), 0);
		strcpy(
				buf,
				"<styleUrl>#radar</styleUrl>\n<Polygon>\n<outerBoundaryIs>\n<LinearRing id=\"radar_range\">\0");
		send(*s, buf, strlen(buf), 0);
		strcpy(buf, "<extrude>0</extrude>\n<tessellate>0</tessellate>\n\0");
		send(*s, buf, strlen(buf), 0);
		strcpy(buf,
				"<altitudeMode>clampToGround</altitudeMode>\n<coordinates>\n\0");
		send(*s, buf, strlen(buf), 0);

		sprintf(sql, "SELECT LON, LAT FROM RADAR_RANGE ORDER BY AZIMUTH\0");
		sqlite3_prepare(db, sql, -1, &stmt_f, 0);
		while (sqlite3_step(stmt_f) == SQLITE_ROW) {
			sprintf(buf, "%f,%f,0\n\0", sqlite3_column_double(stmt_f, 0),
					sqlite3_column_double(stmt_f, 1));
			send(*s, buf, strlen(buf), 0);
		}

		strcpy(
				buf,
				"</coordinates>\n</LinearRing>\n</outerBoundaryIs>\n</Polygon>\n</Placemark>\n\0");
		send(*s, buf, strlen(buf), 0);
	}

	sqlite3_finalize(stmt_f);

	strcpy(buf, "</Document></kml>\n\0");
	send(*s, buf, strlen(buf), 0);
	//sqlite3_mutex_leave(db_mux);

	free(sql);
	free(sql_f);
	free(buf);

	return 0;
}

char * get_mime(char *path) { // mime type by file extension
	char *ext = strchr(path, '.');

	if (!ext)
		return NULL;
	if (!strcmp(ext, ".html") || !strcmp(ext, ".htm"))
		return "text/html";
	if (!strcmp(ext, ".jpeg") || !strcmp(ext, ".jpg"))
		return "image/jpeg";
	if (!strcmp(ext, ".png"))
		return "image/png";
	return "text/plain";
}

void http_send_file(int *s, char *path, int length) {
	int b;
	char buf[1024];

	FILE * f = fopen(path, "r");
	if (!f) {
		http_error((int *) s, 403, "Forbidden", "Access denied.");
		return;
	}

	http_header((int *) s, 200, "OK", get_mime(path), length);

	while ((b = fread(buf, 1, sizeof(buf), f)) > 0)
		send(*s, buf, b, 0);

	fclose(f);
}

void *http(void *arg) /* process client http connection thread */
{
	char * buf = malloc(REQBUFLEN);
	int i;
	int *s = (int*) arg;
	char info[128];
	char path[128];
	struct stat statbuf;
	char * method;
	char * url;

	bzero(buf, REQBUFLEN);
	recv(*s, buf, REQBUFLEN, 0);

	method = strtok(buf, " ");
	url = strtok(NULL, " ");

	if (!method || !url) {
		http_error((int *) s, 400, "Bad Request.", NULL);
		goto exit;
	}

	if (strcmp(method, "GET")) {
		sprintf(info, "Method %s is not supported.\0", method);
		http_error((int *) s, 501, "Not Supported.", info);
		goto exit;
	}

	bzero(path, sizeof(path));
	sprintf(path, "%s/htdocs%s\0", get_current_dir_name(), url);

	if (!strcmp(url, "/stat")) {
		list_flights((int *) s);
	} else if (!strcmp(url, "/adsbox.kml")) {
		make_kml((int *) s);
	} else if (stat(path, &statbuf) < 0) {
		sprintf(info, "URL %s was not found.\0", url);
		http_error((int *) s, 404, "Not Found", info);
	} else if (S_ISDIR(statbuf.st_mode)) {
		http_error((int *) s, 403, "Forbidden", "Access denied."); // dir listing not supported
	} else {
		http_send_file((int *) s, path, statbuf.st_size);
	}

	exit: close(*s);
	free(s);
	free(buf);
	pthread_exit(0);
}
