/*
 *	adsbox - ADS-B decoder software
 *	(c) atty 2011-13 romulmsg@mail.ru
 *
 */
#include <stdio.h>
#include <getopt.h>
#include <stdlib.h> 
#include <sys/ioctl.h>
#include <sys/types.h>
#include <pthread.h>
#include <sys/time.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include "misc.h"
#include "http.h"
#include "../sqlite3/sqlite3.h"
#include "db.h"
#include "file.h"
#include "serial.h"
#include "tcpclient.h"

#define SBS_PORT 30003
#define HTTP_PORT 8080

pthread_mutex_t mux;
pthread_cond_t cond_var;
char gdata[BUFLEN];
char grawdata[BUFLEN];
int dmode = 0;
double homelat = 0;
double homelon = 0;
_Bool record = 0;
_Bool play = 0;
_Bool seed_server = 0;
_Bool data_client = 0;
FILE * rd;
int delay = 0;
sqlite3 *db;
//sqlite3_mutex *db_mux;


void *decoder(void *arg) { /* decoder thread */
	unsigned char rawbuf[BUFLEN];
	unsigned char msg[128];
	struct tm *now;
	struct timespec ts;
	char data[2];
	static int msqid;
	struct msgbuf recvbuf;
	int ret;

	msqid = msgget(msqkey, 0666 | IPC_CREAT);
	if (msqid == -1)
		panic("msgget()");

	bzero(msg, sizeof(msg));
	while (1) {
		bzero(rawbuf, BUFLEN);

		bzero(recvbuf.mtext, sizeof(recvbuf.mtext));
		msgrcv(msqid, &recvbuf, sizeof(recvbuf.mtext), 0, 0);
		sscanf(recvbuf.mtext, "%s", rawbuf);

		// strip time mark if present. Thanks Petr Kouda
		if (strlen(rawbuf) == 28 || strlen(rawbuf) == 42)
			sprintf(rawbuf, "*%s\0", rawbuf + 13);

		if (strlen(rawbuf) == 16 || strlen(rawbuf) == 30) {
			bzero(grawdata, BUFLEN);
			sprintf(grawdata, "%s\n", rawbuf); // raw data to seed
			if (record) {
				clock_gettime(CLOCK_REALTIME, &ts);
				now = localtime(&ts.tv_sec);
				fprintf(rd, "%02d:%02d:%02d.%03d %02d.%02d.%04d %s\n",
						now->tm_hour, now->tm_min, now->tm_sec, ts.tv_nsec
								/ 1000000, now->tm_mday, now->tm_mon + 1,
						now->tm_year + 1900, rawbuf);
				fflush(rd);
			}
			if ((strlen(rawbuf) == 16 && decode_message56(rawbuf, msg))
					|| (strlen(rawbuf) == 30 && decode_message(rawbuf, msg))) {
				pthread_mutex_lock(&mux);
				bzero(gdata, BUFLEN);
				memcpy(gdata, msg, strlen(msg));
				pthread_cond_broadcast(&cond_var);
				pthread_mutex_unlock(&mux);
				bzero(msg, sizeof(msg));
			}
		}
	}
	pthread_exit(0);
}

void *servlet(void *arg) /* process client connection thread. SBS data seed */
{
	char buf[2];
	char cdata[BUFLEN];
	struct timespec to;

	int *s = (int*) arg;

	while (recv(*s, buf, 1, MSG_DONTWAIT)) {
		pthread_mutex_lock(&mux);
		to.tv_sec = time(NULL) + 1;
		to.tv_nsec = 0;
		pthread_cond_timedwait(&cond_var, &mux, &to);
		if (strncmp(cdata, gdata, strlen(gdata))) {
			bzero(cdata, BUFLEN);
			memcpy(cdata, gdata, strlen(gdata));
			send(*s, cdata, strlen(cdata), 0);
		}
		pthread_mutex_unlock(&mux);
	}
	close(*s);
	free(s);
	pthread_exit(0);
}

void *seed(void *arg) /* process client connection thread. Raw data seed */
{
	char buf[2];
	char cdata[BUFLEN];
	struct timespec to;

	int *s = (int*) arg;

	while (recv(*s, buf, 1, MSG_DONTWAIT)) {
		pthread_mutex_lock(&mux);
		to.tv_sec = time(NULL) + 1;
		to.tv_nsec = 0;
		pthread_cond_timedwait(&cond_var, &mux, &to);
		if (strncmp(cdata, grawdata, strlen(grawdata))) {
			bzero(cdata, BUFLEN);
			memcpy(cdata, grawdata, strlen(grawdata));
			send(*s, cdata, strlen(cdata), 0);
		}
		pthread_mutex_unlock(&mux);
	}
	close(*s);
	free(s);
	pthread_exit(0);
}

void usage(void) {
	printf("adsbox [options]\n\t-s, --serial input serial device\n");
	printf(
			"\t--baud serial device baudrate (valid rates are: 115200, 230400, 460800,\n");
	printf("\t\t500000, 576000, 921600, 1000000, 1152000, 1500000,\n");
	printf("\t\t2000000, 2500000, 3000000, 3500000, 4000000)\n");
	printf("\t--bits serial device bits (7 or 8)\n");
	printf(
			"\t-d, --daemon be a daemon\n\t-l, --lat our latitude\n\t-g, --lon our longitude\n");
	printf("\t-r, --record [file] record ADS-B data to file\n");
	printf("\t-p, --play [file] play recordered file\n");
	printf("\t-x, --rate [n] play n times faster\n");
	printf("\t-z, --snooze [n] snooze for n seconds before play\n");
	printf("\t--seed seeding raw ascii data on tcp port (default 30004)\n");
	printf("\t--seed-port [port] seed raw ascii data on this tcp port\n");
	printf("\t--avr-server [IP] connect to avrascii data server\n");
	printf(
			"\t--avr-server-port [port] connect to avrascii data server port (default 30040)\n");
	printf("\t-h, --help this help\n");
	exit(0);
}

int main(int ac, char *av[]) {
	extern char * optarg;
	static struct option long_options[] = {
			{ "daemon", no_argument, NULL, 'd' }, { "serial",
					required_argument, NULL, 's' }, { "help", no_argument,
					NULL, 'h' }, { "lat", required_argument, NULL, 'l' }, {
					"lon", required_argument, NULL, 'g' }, { "record",
					required_argument, NULL, 'r' }, { "play",
					required_argument, NULL, 'p' }, { "rate",
					required_argument, NULL, 'x' }, { "snooze",
					required_argument, NULL, 'z' }, { "baud",
					required_argument, NULL, 1 }, { "bits", required_argument,
					NULL, 2 }, { "seed", no_argument, NULL, 3 }, { "seed-port",
					required_argument, NULL, 4 }, { "avr-server",
					required_argument, NULL, 5 }, { "avr-server-port",
					required_argument, NULL, 6 }, { 0, 0, 0, 0 } };
	int option_index;
	char s[16] = "\0";
	struct sockaddr_in addr, http_addr;
	int sd, hd, sdd = 0, max_d, en = 1;
	int pid, sid;
	char fname[64];
	float playrate = 1;
	long int baudrate = SERIAL_BAUDRATE;
	long int bits = SERIAL_BITS;
	int c, n;
	fd_set input;
	char *zErrMsg = 0;
	int rc;
	int seed_port = AVR_SERVER_PORT;
	char avr_server_IP[64];
	int avr_server_port = AVR_SERVER_PORT;
	pthread_t child;

	pthread_mutex_init(&mux, NULL);
	bzero(fname, 64);

	DEBUG("adsbox ADS-B decoder version %s.\n", VERSION);

	while ((c = getopt_long(ac, av, "ds:hl:g:r:p:x:z:", long_options,
			&option_index)) != -1) {
		switch (c) {
		case 1:
			if (!strcmp(optarg, "115200"))
				baudrate = B115200;
			else if (!strcmp(optarg, "230400"))
				baudrate = B230400;
			else if (!strcmp(optarg, "460800"))
				baudrate = B460800;
			else if (!strcmp(optarg, "500000"))
				baudrate = B500000;
			else if (!strcmp(optarg, "576000"))
				baudrate = B576000;
			else if (!strcmp(optarg, "921600"))
				baudrate = B921600;
			else if (!strcmp(optarg, "1000000"))
				baudrate = B1000000;
			else if (!strcmp(optarg, "1152000"))
				baudrate = B1152000;
			else if (!strcmp(optarg, "1500000"))
				baudrate = B1500000;
			else if (!strcmp(optarg, "2000000"))
				baudrate = B2000000;
			else if (!strcmp(optarg, "2500000"))
				baudrate = B2500000;
			else if (!strcmp(optarg, "3000000"))
				baudrate = B3000000;
			else if (!strcmp(optarg, "3500000"))
				baudrate = B3500000;
			else if (!strcmp(optarg, "4000000"))
				baudrate = B4000000;
			else {
				DEBUG("%s - illegal baudrate.\n", optarg);
				exit(1);
			}
			break;
		case 2:
			if (!strcmp(optarg, "7"))
				bits = CS7;
			else if (!strcmp(optarg, "8"))
				bits = CS8;
			else {
				DEBUG("%s - illegal bits option.\n", optarg);
				exit(1);
			}
		case 3:
			seed_server = 1; // run seed server
			break;
		case 4:
			seed_port = atoi(optarg);
			break;
		case 5:
			strcpy(avr_server_IP, optarg);
			data_client = 1; // connect to avrascii data server
			break;
		case 6:
			avr_server_port = atoi(optarg);
			break;
		case 'd':
			dmode = 1;
			break;
		case 's':
			strncpy(s, optarg, 16);
			break;
		case 'l':
			homelat = atof(optarg);
			break;
		case 'g':
			homelon = atof(optarg);
			break;
		case 'r':
			record = 1;
			play = !record;
			strncpy(fname, optarg, 64);
			break;
		case 'p':
			play = 1;
			record = !play;
			strncpy(fname, optarg, 64);
			break;
		case 'x':
			playrate = atof(optarg);
			break;
		case 'z':
			delay = atoi(optarg);
			break;
		case 'h':
			usage();
			break;
		case '?':
		default:
			DEBUG("Try --help for more info.\n");
			exit(1);
		}
	}

	if ((homelat != 0) && (homelon != 0))
		DEBUG("Your lat %2.5f lon %3.5f\n", homelat, homelon);

	/*
	 * Dealing with sqlite database.
	 */

	/*db_mux = sqlite3_mutex_alloc(SQLITE_MUTEX_RECURSIVE);
	 if (db_mux == NULL) {
	 printf("Can't allocate sqlite mutex.\n");
	 return 1;
	 }*/

	rc = sqlite3_open(":memory:", &db);
	if (rc) {
		DEBUG("Can't open database: %s\n", sqlite3_errmsg((sqlite3 *) &db));
		exit(1);
	}

	if (db_init(db))
		exit(1);
	/*
	 * Done with sqlite.
	 */
	signal(SIGPIPE, SIG_IGN); // ignoring SIGPIPE causing unexpected process kill

	if (dmode) { /* daemonize */
		pid = fork();
		if (pid < 0)
			panic("fork()");

		if (pid > 0)
			exit(0);

		umask(0);

		sid = setsid();

		if (sid < 0)
			panic("setsid()");

		close(STDIN_FILENO);
		close(STDOUT_FILENO);
		close(STDERR_FILENO);
	}

	/*
	 * Start ADS-B decoder thread here
	 */
	pthread_create(&child, NULL, &decoder, NULL); /* start decoder thread */
	pthread_detach(child);

	if (record) {
		if ((rd = fopen(fname, "w+")) == NULL) {
			DEBUG("Unable to open %s\n", fname);
			exit(1);
		}
		DEBUG("Recording data to %s\n", fname);
	}

	if (play) {
		file_info * fi = malloc(sizeof(file_info));
		strcpy(fi->name, fname);
		fi->playrate = playrate;
		fi->delay = delay;
		printf("Playing data from %s file\n", fname);
		pthread_create(&child, NULL, &player_thread, fi); /* start player thread */
		pthread_detach(child);
	}

	if (data_client) {
		tcp_info *ti = malloc(sizeof(tcp_info));
		strcpy(ti->addr, avr_server_IP);
		ti->port = avr_server_port;
		pthread_create(&child, NULL, &tcp_client, ti); /* start tcp client thread */
		pthread_detach(child);
	}

	/* Serial device thread*/
	if (strlen(s)) {
		serial_info *si = malloc(sizeof(serial_info));
		strcpy(si->device, s);
		si->baudrate = baudrate;
		si->bits = bits;
		pthread_create(&child, NULL, &serial_thread, si); /* start serial port thread */
		pthread_detach(child);
	}

	//char msg[256];
	//decode_message("8D4000FB5875444148D9A70B023A", msg);
	//exit(0);

	/*--- create sbs-1 data socket ---*/
	sd = socket(AF_INET, SOCK_STREAM, 0);
	if (sd < 0)
		panic("sbs-1 socket()");
	if (setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &en, sizeof(en)) != 0)
		panic("sbs-1 setsockopt()");

	/* non-blocking socket */
	if (ioctl(sd, FIONBIO, &en) < 0)
		panic("sbs-1 ioctl()");

	/*--- bind  ---*/
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(SBS_PORT);
	addr.sin_addr.s_addr = htonl(INADDR_ANY); /* any interface */
	if (bind(sd, (struct sockaddr*) &addr, sizeof(addr)) != 0)
		panic("sbs-1 bind()");
	if (listen(sd, 10) != 0)
		panic("sbs-1 listen()")
	DEBUG("Run SBS-1 data server on port %i\n", SBS_PORT);

	/*--- create http server socket ---*/
	hd = socket(AF_INET, SOCK_STREAM, 0);
	if (hd < 0)
		panic("http socket()");
	if (setsockopt(hd, SOL_SOCKET, SO_REUSEADDR, &en, sizeof(en)) != 0)
		panic("http setsockopt()");

	/* non-blocking socket */
	if (ioctl(hd, FIONBIO, &en) < 0)
		panic("http ioctl()");

	/*--- bind http server  ---*/
	memset(&http_addr, 0, sizeof(http_addr));
	http_addr.sin_family = AF_INET;
	http_addr.sin_port = htons(HTTP_PORT);
	http_addr.sin_addr.s_addr = htonl(INADDR_ANY); /* any interface */
	if (bind(hd, (struct sockaddr*) &http_addr, sizeof(http_addr)) != 0)
		panic("http bind()");
	if (listen(hd, 10) != 0)
		panic("http listen()")
	DEBUG("Run HTTP server on port %i\n", HTTP_PORT);

	/*--- create seed server socket ---*/
	if (seed_server == 1) {
		sdd = socket(AF_INET, SOCK_STREAM, 0);
		if (sdd < 0)
			panic("seed socket()");
		if (setsockopt(sdd, SOL_SOCKET, SO_REUSEADDR, &en, sizeof(en)) != 0)
			panic("seed setsockopt()");

		/* non-blocking socket */
		if (ioctl(sdd, FIONBIO, &en) < 0)
			panic("seed ioctl()");

		/*--- bind  ---*/
		memset(&addr, 0, sizeof(addr));
		addr.sin_family = AF_INET;
		addr.sin_port = htons(seed_port);
		addr.sin_addr.s_addr = htonl(INADDR_ANY); /* any interface */
		if (bind(sdd, (struct sockaddr*) &addr, sizeof(addr)) != 0)
			panic("seed bind()");
		if (listen(sdd, 10) != 0)
			panic("seed listen()")
		DEBUG("Seeding raw ascii data on TCP port %i\n", seed_port);
	}

	//max_d = (sd > hd ? sd : hd) + 1;
	max_d = sd;
	if (hd > max_d)
		max_d = hd;
	if ((seed_server == 1) && (sdd > max_d))
		max_d = sdd;
	max_d++;

	/*--- begin waiting for connections ---*/
	while (1) { /* process incoming clients */
		FD_ZERO(&input);
		FD_SET(sd, &input);
		FD_SET(hd, &input);
		if (seed_server == 1)
			FD_SET(sdd, &input);

		n = select(max_d, &input, NULL, NULL, 0);
		if (n > 0) {
			if (FD_ISSET(sd, &input)) {
				int *sd1 = (int*) malloc(sizeof(int));
				*sd1 = accept(sd, 0, 0); // accept connection
				fprintf (stdout, "client connected.");
				fflush(stdout);
				if (pthread_create(&child, NULL, &servlet, (void *) sd1)) {
					DEBUG("pthread_create decoder error\n");
					free(sd1);
				}
				pthread_detach(child);
			}
			if (FD_ISSET(hd, &input)) {
				int *hd1 = (int*) malloc(sizeof(int));
				*hd1 = accept(hd, 0, 0); // accept connection
				if (pthread_create(&child, NULL, &http, (void *) hd1)) {
					DEBUG("pthread_create http error\n");
					free(hd1);
				}
				pthread_detach(child);
			}
			if (seed_server && FD_ISSET(sdd, &input)) {
				int *sdd1 = (int*) malloc(sizeof(int));
				*sdd1 = accept(sdd, 0, 0); // accept connection
				if (pthread_create(&child, NULL, &seed, (void *) sdd1)) {
					DEBUG("pthread_create seed_server error\n");
					free(sdd1);
				}
				pthread_detach(child);
			}
		}
	}
	exit(0);
}
