/*
 *	adsbox - ADS-B decoder software
 *	(c) atty 2011-13 romulmsg@mail.ru
 *
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <fcntl.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include "misc.h"
#include "serial.h"

extern char grawdata[BUFLEN];

void *serial_thread(void *arg) {
	serial_info *si = (serial_info*) arg;
	int fd;
	unsigned char rawbuf[BUFLEN];
	struct termios oldtio, newtio;
	static int msqid;
	struct msgbuf sndbuf, *msgp;
	char data[2];

	DEBUG("Using %s as input serial device.\n", si->device);

	fd = open(si->device, O_RDONLY | O_NOCTTY);
	if (fd < 0) {
		DEBUG("Unable to open device %s\n", si->device);
		free(si);
		pthread_exit(0);
	}

	tcgetattr(fd, &oldtio); // TODO restore old termio

	bzero(&newtio, sizeof(newtio));
	newtio.c_cflag &= ~PARENB;
	newtio.c_cflag &= ~CSTOPB;
	newtio.c_cflag &= ~CSIZE;
	newtio.c_cflag |= si->baudrate | si->bits | CLOCAL | CREAD;
	newtio.c_iflag = IGNPAR | ICRNL; // ignore parity error
	newtio.c_oflag = 0;

	//newtio.c_lflag = ICANON; // unfortunately

	newtio.c_cc[VTIME] = 0;
	newtio.c_cc[VMIN] = 1;
	newtio.c_cc[VEOL] = 0;
	newtio.c_cc[VEOL2] = 0;

	tcflush(fd, TCIFLUSH);
	tcsetattr(fd, TCSANOW, &newtio);

	msqid = msgget(msqkey, 0666 | IPC_CREAT);
	msgp = &sndbuf;
	msgp->mtype = 1;

	//bzero(msg, sizeof(msg));
	while (1) {
		bzero(rawbuf, BUFLEN);
		while (strstr(rawbuf, ";") == NULL) {
			read(fd, data, 1);
			if (data[0] == '*' || data[0] == '@')
				bzero(rawbuf, BUFLEN);
			strncat(rawbuf, data, 1);
		}

		// strip time mark if present. Thanks Petr Kouda
		if (strlen(rawbuf) == 28 || strlen(rawbuf) == 42)
			sprintf(rawbuf, "*%s\0", rawbuf + 13);

		//	DEBUG("%i -%s-\n", strlen(rawbuf), rawbuf);
		if (strlen(rawbuf) == 16 || strlen(rawbuf) == 30) {
			bzero(sndbuf.mtext, sizeof(sndbuf.mtext));
			sprintf(sndbuf.mtext, "%s", rawbuf);
			if (msgsnd(msqid, (const void *) msgp, sizeof(sndbuf.mtext),
					IPC_NOWAIT) == -1) {
				DEBUG("player_thread(): msgsend()\n");
				free(si);
				pthread_exit(0);
			}

			bzero(grawdata, BUFLEN);
			memcpy(grawdata, sndbuf.mtext, strlen(sndbuf.mtext)); // raw data to seed
		}
	}

	free(si);
	pthread_exit(0);
}
